//{{NO_DEPENDENCIES}}
// Microsoft Developer Studio generated include file.
// Used by npspy.rc
//
#define IDC_STATIC                      -1
#define IDD_DIALOG_OPTIONS              101
#define IDI_ICON1                       102
#define IDI_ICON_APP                    102
#define ID_DIALOG_CHOOSEDIR             102
#define IDD_DIALOG_PAUSE                103
#define ID_ICON_DRIVEFLOPPY             103
#define IDD_PAGE_GENERAL                104
#define ID_ICON_DRIVEHARD               104
#define IDD_PAGE_FILTER                 105
#define ID_ICON_DRIVENETWORK            105
#define IDD_PAGE_LOG                    106
#define ID_ICON_DRIVECDROM              106
#define IDD_PAGE_ADVANCED               107
#define ID_ICON_DRIVERAM                107
#define ID_ICON_FOLDERCLOSED            108
#define ID_ICON_FOLDEROPEN              109
#define ID_ICON_OPENSELECT              110
#define IDD_MAIN                        308
#define IDD_DIALOG_MAIN                 308
#define IDC_MAIN_OUTPUT                 1001
#define IDC_CHECK_NPN_VERSION           1002
#define IDC_CHECK_NPN_GETURLNOTIFY      1003
#define ID_EDIT_DIR                     1003
#define IDC_CHECK_NPNGETURL             1004
#define IDC_CHECK_NPN_GETURL            1004
#define ID_LIST_DIR                     1004
#define IDC_CHECK_NPN_POSTURLNOTIFY     1005
#define ID_COMBO_DIR                    1005
#define IDC_CHECK_NPN_POSTURL           1006
#define ID_LISTTEMP_DIR                 1006
#define IDC_CHECK_NPN_REQUESTREAD       1007
#define IDC_CHECK_NPN_NEWSTREAM         1008
#define IDC_CHECK_NPN_WRITE             1009
#define IDC_CHECK_NPN_DESTROYSTREAM     1010
#define IDC_CHECK_NPN_STATUS            1011
#define IDC_CHECK_NPN_ESERAGENT         1012
#define IDC_CHECK_NPN_MEMALLOC          1013
#define IDC_CHECK_NPN_MEMFREE           1014
#define IDC_CHECK_NPN_MEMFLUSH          1015
#define IDC_CHECK_NPN_RELOADPLUGINS     1016
#define IDC_CHECK_NPNGETJAVAENV         1017
#define IDC_CHECK_NPN_GETJAVAENV        1017
#define IDC_CHECK_NPN_GETJAVAPEER       1018
#define IDC_CHECK_NPN_GETVALUE          1019
#define IDC_CHECK_NPN_SETVALUE          1020
#define IDC_CHECK_NPN_INVALIDATERECT    1021
#define IDC_CHECK_NPN_INVALIDATEREGION  1021
#define IDC_CHECK_NPNINVALIDATEREGION   1022
#define IDC_CHECK_NPN_FORCEREDRAW       1023
#define IDC_CHECK_NPP_NEW               1024
#define IDC_CHECK_NPP_DESTROY           1025
#define IDC_CHECK_NPP_SETWINDOW         1026
#define IDC_CHECK_NPP_NEWSTREAM         1027
#define IDC_CHECK_NPP_DESTROYSTREAM     1028
#define IDC_CHECK_NPP_STREAMASFILE      1029
#define IDC_CHECK_NPP_WRITEREADY        1030
#define IDC_CHECK_NPP_WRITE             1031
#define IDC_CHECK_NPP_PRINT             1032
#define IDC_CHECK_NPP_HANDLEEVENT       1033
#define IDC_CHECK_NPP_URLNOTIFY         1034
#define IDC_CHECK_NPP_GETJAVACLASS      1035
#define IDC_CHECK_NPP_GETVALUE          1036
#define IDC_CHECK_NPP_SETVALUE          1037
#define IDC_BUTTON_CHECKALL             1038
#define IDC_BUTTON_CLEARALL             1039
#define IDC_CHECK_TOCONSOLE             1041
#define IDC_CHECK_TOWINDOW              1042
#define IDC_CHECK_TOFILE                1043
#define IDC_CHECK_SPALID                1044
#define IDC_EDIT_FILE                   1047
#define IDC_BUTTON_CHOOSEDIR            1048
#define IDC_CHECK_MUTE                  2019
#define IDC_CHECK_ONTOP                 2020
#define IDC_BUTTON_OPTIONS              2023
#define IDC_BUTTON_CLEAR                2024
#define IDC_OUTPUT                      2026

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NEXT_RESOURCE_VALUE        108
#define _APS_NEXT_COMMAND_VALUE         40001
#define _APS_NEXT_CONTROL_VALUE         1049
#define _APS_NEXT_SYMED_VALUE           101
#endif
#endif
